/* File:      cov_mat.h
   Date:      5/19/03  (code originally written fall 2001)
   Author:    C. Paciorek
   Purpose:   header file for cov_mat.C
*/

const double TOL=.0000000001;  /* 1X10^-10; 1X10^-9 might be best; at one point I also used 1X10^-10 */
/* flags for warning whether gen_chol zeroes out any columns */
const int WARN=1;  
const int NO_WARN=0;

class covariate_info;
/* covariate information object */
extern covariate_info* covariates;

extern "C" void smbfct_(int* n,int* xadj,int* adjncy,int* perm,int* invp,int* xlnz,int* maxlnz,int* xnzsub,int* nzsub,int* maxsub,int* rchlnk,int* mrglnk,int* marker); /* Fortran routine for symbolic factorization, which creates the sparse matrix representation  */

extern "C" void gsfct_(int* n,int* xlnz,double* lnz,int* xnzsub,int* nzsub,double *diag,int* link,int* first,double *tmp,double* eps,int* war);  /* Fortran routine for numeric factorization */

/* Note that in the various classes, L indicates Cholesky decomposition, C a covariance matrix and R a correlation matrix; the exception is sparse_cov_mat_sum which used objects labelled R for covariance matrices */ 

class sparse_cov_mat{
/* higher level structure for sparse covariance matrices */
public:
  sparse_cov_mat(int size_in,int num_covars_in,double* smoothness_in);
  ~sparse_cov_mat(void);
  void sym_chol(void);                    /* symbolic factorization -- sets up sparse structure */
  void gen_chol(int warn);                /* numeric factorization */
  int Csolve(double* b, double* output);  /* calcs C^{-1}b */
  void Rmult(double* b, double* out);     /* multiplies R by vector b */
  void Lmult(double* b, double* out);     /* multiplies L times vector b */
  void Ltmult(double* b, double* out);    /* multiplies L^T times vector b */
  void Ltmult_old(double* b, double* out);       /* multiplies L^T times vector b using stored version of Lt */
  void Lcheck(double* newvec,double* oldvec);
  double det_log_squ_root(void);          /* calc sqrt of log(determinant) via sum of log of Cholesky diagonals */
  void fill(double* input);        /* fill data structure for matrix */
  void update(int accepted);       /* update current and stored values depending on whether proposed values are accepted */
  double *lnz_R;                   /* nonzero off-diag values by column, then by row;  in public because need to pass it through data to data cov */
  double* lnz_R_stored;            /* in public because need to pass it through data to data cov */
  int size;
  mat_pp<double>* euc_distance;    /* angular distance between points for each element of cov matrix */
protected:                 /* I think these are protected so they can be used by the children classes */
  int num_covars;
  double* smoothness;       /* smoothness parm of Matern correlation */
  double* diag_R;           /* diag values of correlation */
  double* diag_L;           /* diag values of Cholesky */
  double* lnz_L;            /* non-zero off-diag values of Cholesky */

  /* stored values for restoration if proposal not accepted */
  double* lnz_L_stored;     
  double* diag_L_stored;
  double* diag_R_stored;

  int maxlnz;               /* length of lnz */
  int maxsub;               /* length of nzsub */
  int *nzsub;               /* row indices of the off-diagonal values, with repeats eliminated */
  int *xlnz;                /* vector of pointers (1,...,n+1) of where the elements of a column start -- xlnz[0] is where the elements of col 1 start; xlnz[1] is where col 2 starts,  */
  int *xnzsub;              /* vector of pointers (1,...,n+1) of start of column in nzsub -- xnzsub[0] is where the row indices of column 1 start; xnzsub[1] for column 2, etc. */
  /* adjacency and permutation information are part of covariate_info class -- cov_mat is a friend of covariates */
};

class cov_mat_calc{
/* structure for doing straightforward calculations for a matrix loaded from a vector */
public:
  cov_mat_calc(int size_in);
  ~cov_mat_calc(void);
  void sym_chol(void);                /* symbolic factorization -- sets up sparse structure */
  void gen_chol(int warn);            /* numeric factorization */
  int Csolve(double* b, double* output);  /* calcs C^{-1}b */
  double det_log_squ_root(void);       /* calc sqrt of log(determinant) via sum of log od Cholesky diagonals */
  void fill(double* input);            /* fills structure from a vector input */
private:
  int size;

  /* see comments in sparse_cov_mat */
  double *lnz_C;            
   double *diag_L;
  double* lnz_L; 
  double* diag_C;          

  /* stored values for restoration if proposal rejected */
  double* lnz_L_stored;   
  double* lnz_C_stored;  
  double *diag_L_stored;
  double *diag_C_stored;

  int maxlnz;               /* length of lnz */
  int maxsub;               /* length of nzsub */
  int *nzsub;               /* row indices of the off-diags, with repeats eliminated */
  int *xlnz;                /* vector of pointers (1,...,n+1) of where the elements of a column start -- xlnz[0] is where the elements of col 1 start; xlnz[1] is where col 2 starts,  */
  int *xnzsub;              /* vector of pointers (1,...,n+1) of start of column in nzsub -- xnzsub[0] is where the row indices of column 1 start; xnzsub[1] for column 2, etc. */

  /* adjacency information */
  int* xadj;  
  int* adjncy;
};

/* class for sum of two covariance matrices; used for C_f+C_y where f indicates prior function covariance and y indicates data covariance */
class sparse_cov_mat_sum : public sparse_cov_mat{
 public:
  sparse_cov_mat_sum(int size_in,double* fun_lnz_R_in,double* fun_lnz_R_stored_in,double* fun_sigma_in,double* fun_sigma_stored_in,double* data_sigma_in,double* fun_in,double* num_trials_in);
  ~sparse_cov_mat_sum(void);
  void fill();
  void fill_old();
  double calc_df(void);     /* calculate degrees of freedom for smoother as trace(C_f(C_f+C_y)^-1)  */
 private:
  double* fun;              /* function values */
  double* num_trials;
  double* fun_sigma;        /* prior sd for function */
  double* fun_sigma_stored; /* stored version needed for Hastings calc */
  double* data_sigma;       /* noise sd */
  double* fun_lnz_R;        /* off-diags of prior correlation matrix */
  double* fun_lnz_R_stored; /* stored version needed for Hastings calc */
  mat_pp<double>* Cfmat;    /* representation of prior covariance matrix; used in df calc */
};

/* child class for stationary covariance matrices */
class sparse_cov_mat_stat : public sparse_cov_mat{
public:
  sparse_cov_mat_stat(int size_in,int num_covars_in, double** kappa_in,double** gamma_in,double* smoothness_in);
  ~sparse_cov_mat_stat(void);  
  void fill(int first);            /* calculate values in matrix */
  double fillij(int index);        /* calculate ijth element of matrix */
private:
  double** kappa;                  /* vector of eigenvals of kernel matrices */
  double** gamma;                  /* vector of angles of kernel matrices */
  double* kernel;                  /* vector representation of individual kernel matrix */
  cov_mat_calc* kernel_mat;        /* holds matrix representation of individual kernel matrix for matrix calculations */
  double* vec;
};

/* child class for nonstationary covariance matrices */
class sparse_cov_mat_nonstat : public sparse_cov_mat{
public:
  sparse_cov_mat_nonstat(int size_in,int num_covars_in, double** eigenval_in,double** eigenval_in,double* smoothness_in);
  ~sparse_cov_mat_nonstat(void);
  void update(int accepted);             /* update current and stored values depending on whether proposed values are accepted */
  void fill(int which, int first);       /* calculate values in matrix */
  double fillij(int i, int j,int index); /* calculate ijth element of matrix */
private:
  double** eigenval;                     /* vector of pointers to eigenvalues of kernel matrices */
  double** eigenvec;                     /* vector of pointers to eigenvector coordinates of kernel matrices */

  mat_pp<double>* kernel;                /* set of vector representations of kernel matrices */
  mat_pp<double>* eigenvecs;
  mat_pp<double>* eigenvecs_stored;      /* saved eigenvectors for when proposing other parts of covariance structure */

  double* lower;
  double* vec;
  double* kernel_sum;                    /* holds vector representation of summed kernel matrices during nonstationary correlation calculation */

  cov_mat_calc* kernel_sum_mat;          /* holds matrix representation of summed kernel matrices during nonstationary correlation calculation */

};










