/* File:      covariates.C
   Date:      3/20/02  (written fall 2001)
   Author:    C. Paciorek
   Purpose:   code for dealing with covariate information
   Compile:   g++ -c  -Wall -ansi
   Usage:     N/A
   Arguments: N/A
   Requires:  see just below
   Reference: George & Liu 1981:  - this source has the algorithms and storage format of the adjacency info  */

/*
@Book{Geor:Liu:1981,
Author={George, A. and J.W.-H. Liu},
Title={Computer solution of large sparse positive definite systems},
Year=1981,
Address={Englewood Cliffs, New Jersey},
Publisher={Prentice-Hall, Inc.},
Pages=324
}
*/

#include "base.h"
#include "covariates.h"
#include "permtools.h"


covariate_info::covariate_info(string covar_fn_in,string perm_fn_in,int num_trainpoints_in,int num_testpoints_in,int num_covars_in){

  num_covars=num_covars_in;
  num_trainpoints=num_trainpoints_in;
  num_testpoints=num_testpoints_in;
  num_datapoints=num_trainpoints+num_testpoints;
  num_angles=(int)num_covars*(num_covars-1.0)/2.0;
  num_eigenvecs=(int)(pow(num_covars,2.0)+num_covars-2.0)/2.0;
 
  covars=new mat_pp<double>(num_datapoints,num_covars);
  if(num_covars==1){
    covars->read(covar_fn_in,BY_ROW,RAGGED);  
  }
  else{  // covariates should be one observation per line
    covars->read(covar_fn_in,BY_ROW,MATRIX);  
  }

  scale_covars();   // scale covariates to (0,1)

  fill_adj();     // creates adjacency structure used in symbolic factorization to create sparse structure - though the matrices I use are not sparse

  // read in permutation and inverse permutation
  mat_pp<int> tmp(1,num_datapoints);

  tmp.read(perm_fn_in,BY_ROW,RAGGED);
  perm=tmp.getvec(0,ROW);

  invp=new int[num_datapoints];
  invrse(&num_datapoints,perm,invp,PERMSTART1);

  // note that call to calc_datadata_distance is done in cov_mat.C since distances are stored in a format compatible with the sparse matrix structure
}


covariate_info::~covariate_info(void){
  delete covars;
  delete [] covar_max;
  delete [] covar_min;

  delete [] adjncy_train;
  delete [] xadj_train;
  if(num_testpoints>0){
    delete [] adjncy_full;
    delete [] xadj_full;
  }
  
  delete [] perm;
  delete [] invp;
}

void covariate_info::fill_adj(void){
  /* determines pairs of covariates for which correlations not  zero;
     fills adjacency structures based on this -- adjacency structures use indices 1,...,N (Fortran style)
     to indicate which covariates are neighbors of which covariates
     but neighbor_ptr uses 0,...,N-1 */

  int i, j, newi;
  int* tmp_adjncy;

  // adjacency info for training covariates
  tmp_adjncy=new int[num_trainpoints*num_trainpoints+1];              // adjacency vector in George & Liu form - Fortran style indexing - temporary since not sure how long it should be
  xadj_train=new int[num_trainpoints+1];                           // for each covariate, pointer to start of neighbors in tmp_adjncy/adjncy - Fortran style indexing
  
  adjncy_len_train=0;
  for(i=0;i<num_trainpoints;i++){
    newi=1;

    for(j=0;j<num_trainpoints;j++){ 
      if(i!=j){  // added 1/17/03; doesn't seem to affect Cholesky data structures; go back to George & Liu to see whether this line should be here
	tmp_adjncy[adjncy_len_train]=j+1;
	if(newi){
	  xadj_train[i]=adjncy_len_train+1;
	  newi=0;
	}
	adjncy_len_train++;
      }
    }
  }
  
  adjncy_len_train++;                                  // this leaves one empty space
  xadj_train[num_trainpoints]=adjncy_len_train;
  adjncy_train=new int[adjncy_len_train];              // make vector to be exactly the right length
  for(i=0;i<adjncy_len_train;i++){
    adjncy_train[i]=tmp_adjncy[i];
  }
  delete [] tmp_adjncy;

  cout << "Number of training points: " << adjncy_len_train << endl;

  // adjacency info for full set of covariates
  if(num_testpoints>0){
    int *tmp_adjncy=new int[num_datapoints*num_datapoints+1];              // adjacency vector in George & Liu form - Fortran style indexing - temporary since not sure how long it should be
    xadj_full=new int[num_datapoints+1];                           // for each covariate, pointer to start of neighbors in tmp_adjncy/adjncy - Fortran style indexing
    
    adjncy_len_full=0;
    for(i=0;i<num_datapoints;i++){
      newi=1;
      for(j=0;j<num_datapoints;j++){ 
	tmp_adjncy[adjncy_len_full]=j+1;
	if(newi){
	  xadj_full[i]=adjncy_len_full+1;
	  newi=0;
	}
	adjncy_len_full++;
      }
    }
    
    adjncy_len_full++;                                  // this leaves one empty space
    xadj_full[num_datapoints]=adjncy_len_full;
    adjncy_full=new int[adjncy_len_full];                    // make vector to be exactly the right length
    for(i=0;i<adjncy_len_full;i++){
      adjncy_full[i]=tmp_adjncy[i];
    }
    delete [] tmp_adjncy;
    
    cout << "Number of total points: " << adjncy_len_full << endl;
  }
}

void covariate_info::choose_adjinfo(int size){
  // choose which adjacency structure is currently active: training set or training plus test set (used in Cholesky calculations)
  if(size==num_trainpoints){
    adjncy=adjncy_train;
    xadj=xadj_train;
  }
  else{
    if(num_testpoints>0 && size==num_datapoints){
      adjncy=adjncy_full;
      xadj=xadj_full;
    }
    else{
      cerr << "Error: size of covariance matrix does not match size of covariate adjacency information.\n";
    }
  }
}

void covariate_info::calc_datadata_distance(int x,int y,double* dist,int permute){
  /* calcs Euclidean distances between covariates x and y */
  int permx=x;  
  int permy=y;
  if(permute==PERMUTE){   // permute covariates if covariance is in permuted form
    permx=perm[x]-1;      // -1 converts to C-style indexing from Fortran-style
    permy=perm[y]-1;
  }
  for(int p=0;p<num_covars;p++){
    // simply Euclidean distance in \Re^{1}; do not use absolute values here since I plug these distances into a quadratic form Mahalanobis distance calculation in cov_mat.C
    dist[p]=covars->vals[permx][p]-covars->vals[permy][p];
  }
}

void covariate_info::scale_covars(void){
  /* scales covariates to be in the range (0,1) */

  int i,p; double diff;
  covar_max=new double[num_covars];
  covar_min=new double[num_covars];
  for(p=0;p<num_covars;p++){
    covar_max[p]=covars->vals[0][p];
    covar_min[p]=covars->vals[0][p];
  }
  // find max and min values
  for(i=1;i<num_datapoints;i++){
    for(p=0;p<num_covars;p++){
      if(covars->vals[i][p]<covar_min[p]){
	covar_min[p]=covars->vals[i][p];
      }
      if(covars->vals[i][p]>covar_max[p]){
	covar_max[p]=covars->vals[i][p];
      }
    }
  }
  // scale covariates to (0,1)
  for(i=0;i<num_datapoints;i++){
    for(p=0;p<num_covars;p++){
      diff=covar_max[p]-covar_min[p];
      covars->vals[i][p]=(covars->vals[i][p]-covar_min[p])/diff;
    }
  }
}

















