/* File:      dataset.C
   Date:      5/19/03 (heavily revised from fall 2001 code)
   Author:    C. Paciorek
   Purpose:   code for storing data, calculating sufficient statistics, calculating likelihood
   Compile:   g++ -c -Wall -ansi
   Usage:     N/A
   Arguments: N/A
   Requires:  see just below
   Reference: written by C. Paciorek
*/

#include "base.h"
#include "dataset.h"
#include "covariates.h"
#include "cov_mat.h"
#include "density.h"             // uses gamma_dens for df
#include "permtools.h"           // access to permutation for permuting data

dataset::dataset(string dat_fn_in,string fun_form_fn_in,double* noise_sd_in,double* noise_sd_stored_in,double df_parm1_in, double df_parm2_in){  

  // zero should be entered in init file if don't want prior on df
  df_parm1=df_parm1_in;
  df_parm2=df_parm2_in;
  
  logprior_df=0.0;  // initialize here so zeroed out if no prior on df
  df=0.0;

  // read in true function values (for simulated data) or train plus test observations (for real data)
  mat_pp<double> tmp(1,covariates->num_datapoints);
  tmp.read(fun_form_fn_in,BY_ROW,RAGGED);
  fun_form=tmp.getvec(0,ROW);

  // read in train observations
  mat_pp<double> tmp2(covariates->num_trainpoints,2);
  tmp2.read(dat_fn_in,BY_ROW,MATRIX);
  Y=tmp2.getvec(1,COL);
  num_trials=tmp2.getvec(0,COL);  // number of Bernoulli trials

  noise_sd=noise_sd_in;
  noise_sd_stored=noise_sd_stored_in;

  
  // permute vectors
  permrv(&covariates->num_datapoints,fun_form,covariates->perm,PERMSTART1);   
  permrv(&covariates->num_trainpoints,Y,covariates->perm,PERMSTART1);    // need to permute data vectors to match parameter vectors

  Yprime=new double[covariates->num_trainpoints];

  initialize_postmean();
}

dataset::~dataset(){
  delete [] Y;                 // allocated in mat_pp when I "getvec"
  delete [] num_trials;        // allocated in mat_pp when I "getvec"
  delete [] Yprime;
  delete [] fun_form;          // allocated in mat_pp when I "getvec"
}

void dataset::ytoyprime(){
  for(int i=0;i<covariates->num_trainpoints;i++){
    Yprime[i]=log(fun[i]/(1-fun[i]))+(Y[i]-num_trials[i]*fun[i])/(num_trials[i]*fun[i]*(1-fun[i]));
  }
}

void dataset::add_info(double* fun_in,double* fun_full_in,double* fun_lnz_R_in,double* fun_lnz_R_stored_in,double* fun_sigma_in,double* fun_sigma_stored_in){
  // df stuff
  cov=new sparse_cov_mat_sum(covariates->num_trainpoints,fun_lnz_R_in,fun_lnz_R_stored_in,fun_sigma_in,fun_sigma_stored_in,noise_sd,fun_in,num_trials);
  cov->fill();
  cov->gen_chol(WARN);
  fun=fun_in;
  fun_test=fun_full_in;
  evaluate_loglik_diff(FUN);                           // calc initial likelihood
  update(ACCEPT,FUN);                                  // update stored likelihood
}

void dataset::update(int accepted,int which){
  // if accepted, stored value becomes current proposed values, if not, restore the stored value as the current value;  depending on parameter being proposed, some portions of likelihood stay fixed

 if(accepted){
   loglik=0.0;
   for(int i=0;i<covariates->num_trainpoints;i++){
     loglik+=Y[i]*log(fun[i]/(1-fun[i]))+num_trials[i]*log(1-fun[i]);
   }
   logprior_df_stored=logprior_df;
   df_stored=df;
   loglik_stored=loglik;
  }   
  else{
    loglik=loglik_stored;
    logprior_df=logprior_df_stored;
    df=df_stored;
  }
  cov->update(accepted);
}


double dataset::evaluate_loglik_diff(int which){
  // calculates loglikelihood, but only portion that changes with the parameter being proposed

  // df stuff
  if(which==ETA){
    cov->fill();
    cov->gen_chol(WARN);
  }

  df=cov->calc_df();
  if(!(zero(df_parm1) || zero(df_parm2))){
    logprior_df=gamma_dens(df,df_parm1,df_parm2);
  }

   loglik=0.0;
   for(int i=0;i<covariates->num_trainpoints;i++){
     loglik+=Y[i]*log(fun[i]/(1-fun[i]))+num_trials[i]*log(1-fun[i]);
   }

   return(loglik-loglik_stored+logprior_df-logprior_df_stored);
}

/////////////////////////////////
// Evaluation Code
/////////////////////////////////

// these calculations are really designed for normal likelihoods, but I'm leaving them in for now

void dataset::initialize_postmean(void){
  postmean = new double[covariates->num_datapoints];
  postmean_final = new double[covariates->num_datapoints];
  for(int i=0;i<covariates->num_datapoints;i++){
    postmean[i]=0.0;
  }
  postmean_count=0;  
}

void dataset::update_postmean(void){
  // updates posterior mean (adding to running total) based on current iterate
  for(int i=0;i<covariates->num_trainpoints;i++){
    postmean[i]+=fun[i];
  }
  if(covariates->num_testpoints){
    for(int i=covariates->num_trainpoints;i<covariates->num_datapoints;i++){
      postmean[i]+=fun_test[i];
    }
  }
  postmean_count++;
}


double* dataset::finalize_postmean(void){
  // calculate average over MCMC iterations to find current posterior mean up to and including current iterate
  for(int i=0;i<covariates->num_datapoints;i++){
    postmean_final[i]=(double)postmean[i]/postmean_count;
  }
  return(postmean_final);
}

double* dataset::calculate_MSE(void){
  // calculates MSE/FVU of training points
  double* output=new double[4];
  double MSE,f2sum,fsum,varf,FVU;
  MSE=0.0;
  f2sum=0.0;
  fsum=0.0;
  for(int i=0;i<covariates->num_trainpoints;i++){
     MSE+=pow(fun_form[i]-postmean_final[i],2.0);
     f2sum+=fun_form[i]*fun_form[i];
     fsum+=fun_form[i];
  }
  MSE/=(double)covariates->num_trainpoints;
  varf=(double)(f2sum-fsum*fsum/(double)covariates->num_trainpoints)/(double)covariates->num_trainpoints;
  FVU=MSE/varf;
  output[0]=MSE; output[1]=FVU;

  // MSE/FVU for test points
  if(covariates->num_testpoints){
    MSE=0.0;
    f2sum=0.0;
    fsum=0.0;
    for(int i=covariates->num_trainpoints;i<covariates->num_datapoints;i++){
      // just assess for test points
      MSE+=pow(fun_form[i]-postmean_final[i],2.0);
      f2sum+=fun_form[i]*fun_form[i];
      fsum+=fun_form[i];
    }
    MSE/=(double)covariates->num_testpoints;
    varf=(double)(f2sum-fsum*fsum/(double)covariates->num_testpoints)/(double)covariates->num_testpoints;   // normalize MSE by variance of function(observation) values
    FVU=MSE/varf;
    output[2]=MSE; output[3]=FVU;
  }
  return(output);
}

void dataset::current_eval(void){
  // calculates values of MSE for current iterate
  current_MSE_train=0.0; current_MSE_test=0.0;
  for(int i=0;i<covariates->num_trainpoints;i++){
    current_MSE_train+=pow(fun_form[i]-fun[i],2.0);
  }
  if(covariates->num_testpoints){
    for(int i=covariates->num_trainpoints;i<covariates->num_datapoints;i++){
      current_MSE_test+=pow(fun_form[i]-fun_test[i],2.0);
    }
  }
  current_MSE_train/=covariates->num_trainpoints;
  current_MSE_test/=covariates->num_testpoints;
}













