
/* File:      regrmod.C
   Date:      5/19/03
   Author:    C. Paciorek
   Purpose:   contains the main function for the Bayesian GP regression model for my dissertation
   Compile:   g++ -c -Wall -ansi
   Usage:     nsgp data_dir num_files num_train num_test num_covariates num_burnin num_iterations stationarity storage_interval write_interval
   Arguments: see below and example.csh 
   Requires:  see just below
   Reference: my dissertation

*/

/* Code files:
   regrmod.C     main
   generic_functions.C   basic functions
   dataset.C     data storage and likelihood calc
   parm.C        parameter class: initialization,storage,updating
   chain.C       stores data and parameters and does MCMC steps
   gasdev.C      random normal generation
   covariates.C  loads covariate info, computes sparse kernel matrix information
   ran1.C        random uniform generation
   transform.C   transformation functions
   density.C     density functions
   mat_pp.h      storage via array of pointers to pointers (arrays)
   cov_mat.C     covariance matrix code (non-discretized version)
   permtools.C   basic permutation functions
   corr.C        correlation functions
   gsfct.f       numeric Cholesky factorization
   smbfct.f      symbolic Cholesky factorization -- sets up sparse structure
   rkbesld.f     modified Bessel function of second kind calculations
*/

#include "base.h"
#include "dataset.h"  // needed for evaluation functions
#include "covariates.h"
#include "chain.h"    

  // argv[1] is the data directory (x,y,fun,init,perm,train(if necessary))
  // argv[2] is the number  indexing the files to be used (of yN.dat,initN.dat,etc.)
  // argv[2] is the number of training data points
  // argv[3] is the number of test data points
  // argv[4] is the number of covariates
  // argv[5] is the number of burn-in iterations
  // argv[6] is number of iterations
  // argv[7] is 0 or `stat' for stationary prior and 1 or `nonstat' for nonstationary
  // argv[8] is the interval between iterate storage (0 if no storage requested)
  // argv[9] is the number of iterates stored before output is written

int* idum;

covariate_info* covariates;    
dataset* data;           // dependent variable (observations) ; called by all the parm classes

int prior;
string data_dir;
string output_fn;

// classes for passing information to the run_model function; these classes and the run_model function are vestiges of a parallel processing implementation
class input_message{
public:
  int num_its;
  int num_burnin;
  int num_train;
  int num_test;
  int num_covars;
  int num_runs;
  int sample;
  int prior;
  int storage_interval;
  int storage_space;
};

class output_message{
public:
  double MSE_train;
  double FVU_train;
  double MSE_test;
  double FVU_test;
  int sample;
};

output_message run_model(input_message input){

  int i;
  
  string tmp,tmp2,covar_fn,perm_fn,dat_fn,init_fn,fun_form_fn,parm_fn,final_fn,log_fn,output_fn,eval_fn,postmean_fn;
  
  output_message output;

  prior=input.prior;

  tmp=itos(input.sample);

  covar_fn=data_dir+"/x"+tmp+".dat";
  fun_form_fn=data_dir+"/fun"+tmp+".dat";
  perm_fn=data_dir+"/perm"+tmp+".dat";
  dat_fn=data_dir+"/y"+tmp+".dat";
  init_fn=data_dir+"/init"+tmp+".dat";
  eval_fn=data_dir+"/eval"+tmp+".dat";
  parm_fn=data_dir+"/parm"+tmp+".dat";
  final_fn=data_dir+"/final"+tmp+".dat";
  log_fn=data_dir+"/log"+tmp+".dat";
  output_fn=data_dir+"/output"+tmp+".dat";
  postmean_fn=data_dir+"/postmean"+tmp+".dat";

  covariates=new covariate_info(covar_fn,perm_fn,input.num_train,input.num_test,input.num_covars);      // read in and set up covariates
  
  chain* this_chain;
    
  int store;
  if(!input.storage_interval){
    store=FALSE;
    input.storage_space=100;  
  }
  else{
    store=TRUE;
  }
  
  this_chain=new chain(log_fn,eval_fn,fun_form_fn,dat_fn,init_fn,final_fn,parm_fn,output_fn,input.num_its,input.storage_interval,input.storage_space);   
  // set up parameters, data, etc.
  
  // I have this_chain be a pointer because otherwise it's deleted after covariates yet covariates->num_covars used in destructor for chain
  
  time_t t1,t2,diff;
  char *timeout;
  
  t1=time(0);
  timeout=ctime(&t1);
  cout << "Process " << input.sample << ": Start at: " << timeout << endl; cout.flush();
  
  for(i=1;i<=input.num_burnin;i++){
      this_chain->do_a_step(i,BURNIN); 
      this_chain->store(i,store,BURNIN);             // store iterates in memory after every iteration
  }
  cout << "Process " << input.sample << ": Finished burn-in period of " << input.num_burnin << " iterations.\n";
  for(i=1;i<=input.num_its;i++){
    this_chain->do_a_step(i+input.num_burnin,NOT_BURNIN); 
    this_chain->store(i+input.num_burnin,store,NOT_BURNIN);             // store iterates in memory after every iteration
  }
  t2=time(0);
  timeout= ctime(&t2);
  diff=t2-t1;
  cout << "Process " << input.sample << ": Time for " << input.num_its << " MCMC iterations: " << diff << ".\n Ended at: " << timeout << endl;cout.flush();
  
  if(store){
    this_chain->report(input.num_its,FINAL);        // final write to output file
  }
  
  double* tmpout;
  
  // calculate posterior mean
  tmpout=data->finalize_postmean();
  ofstream postmeanfile;
  postmeanfile.open(postmean_fn.c_str(),ios::app);
  for(int i=0;i<covariates->num_datapoints;i++){
    postmeanfile << tmpout[i] << " ";
  }
  postmeanfile << endl;
  postmeanfile.close();
  
  // calculate MSE and FVU evaluation criteria
  tmpout=data->calculate_MSE();
  output.MSE_train=tmpout[0];
  output.FVU_train=tmpout[1];
  output.MSE_test=tmpout[2];
  output.FVU_test=tmpout[3];
  delete [] tmpout;
  output.sample=input.sample;
  delete this_chain;
  
  
  delete covariates;

  return(output);
}

int main(int argc, char **argv){

  cout << "Running regression model with: " << endl;
  cout << "Data directory: " << argv[1] << endl;
  cout << "Number of run: " << argv[2] << endl;
  cout << "Number of train points: " << argv[3] << endl;
  cout << "Number of test points: " << argv[4] << endl;
  cout << "Number of covariates: " << argv[5] << endl;
  cout << "Number of burnin iterations: " << argv[6] << endl;
  cout << "Number of sampling iterations: " << argv[7] << endl;
  cout << "Type of covariance: " << argv[8] << endl;
  cout << "Iterate storage interval: " << argv[9] << endl;
  cout << "Write-to-file interval: " << argv[10] << endl;
   
  
  input_message input;
  output_message output;
  
  
  time_t t1, t2, diff;
  char *timeout;
  
  data_dir=argv[1];
  
  input.num_its=atoi(argv[7]);
  input.num_burnin=atoi(argv[6]);
  input.num_train=atoi(argv[3]);
  input.num_test=atoi(argv[4]);
  input.num_covars=atoi(argv[5]);
  input.storage_interval=atoi(argv[9]);
  input.storage_space=atoi(argv[10]);
  
  input.sample=atoi(argv[2]);
  
  string priorcorr_structure=argv[8];
  if(priorcorr_structure=="0" || priorcorr_structure=="stat" ||  priorcorr_structure=="stationary"){
    input.prior=STATIONARY;
    
  }
  else{
    if(priorcorr_structure=="1" || priorcorr_structure=="nonstat" ||  priorcorr_structure=="nonstationary"){
      input.prior=NON_STATIONARY;
    }
    else{
      cerr << "Warning: covariance type " << priorcorr_structure << " not recognized.\n";
    }
  }
  
  t1=time(0);
  timeout=ctime(&t1);
  
  output=run_model(input);

  cout << "Results\tMSE\tFVU\n";
  cout << "Train:\t" << output.MSE_train << "\t" << output.FVU_train << endl;
  if(input.num_test){
    cout << "Test:\t" << output.MSE_test << "\t" << output.FVU_test << endl;
  }

  string tmp=itos(input.sample);
  string eval_fn=data_dir+"/eval"+tmp+".dat";
  ofstream evalfile;
  evalfile.open(eval_fn.c_str(),ios::app);
  evalfile << setprecision(5) << input.sample << " ";
  evalfile << setprecision(5) << output.MSE_train << " ";
  evalfile << setprecision(5) << output.FVU_train << " ";
  evalfile << setprecision(5) << output.MSE_test << " ";
  evalfile << setprecision(5) << output.FVU_test << " ";
  evalfile << endl;
  evalfile.close();
  
  t2=time(0);
  timeout= ctime(&t2);
  diff=t2-t1;
  cout << "Ended at: " << timeout << ". Time used " << diff << endl;cout.flush();
}








